@@ -1,6 +1,7 @@
 ;; service.scm -- Representation of services.
-;; Copyright (C) 2013 Ludovic Courts <ludo@gnu.org>
-;; Copyright (C) 2002, 2003 Wolfgang Jhrling <wolfgang@pro-linux.de>
+;; Copyright (C) 2013, 2014 Ludovic Courtès <ludo@gnu.org>
+;; Copyright (C) 2002, 2003 Wolfgang Järling <wolfgang@pro-linux.de>
+;; Copyright (C) 2014 Alex Sassmannshausen <alex.sassmannshausen@gmail.com>
 ;;
 ;; This file is part of GNU dmd.
 ;;
@@ -59,6 +60,8 @@
             handle-unknown
             make-forkexec-constructor
             make-kill-destructor
+            exec-command
+            fork+exec-command
             make-system-constructor
             make-system-destructor
             make-init.d-service
@@ -227,50 +230,50 @@
 ;; Stop the service, including services that depend on it.  If the
 ;; latter fails, continue anyway.  Return `#f' if it could be stopped.
 (define-method (stop (obj <service>) . args)
-  (if (not (running? obj))
-      (local-output "Service ~a is not running." (canonical-name obj))
-    (if (slot-ref obj 'stop-delay?)
-	(begin
-	  (slot-set! obj 'waiting-for-termination? #t)
-	  (local-output "Service ~a pending to be stopped."
-			(canonical-name obj)))
-      (begin
-	;; Stop services that depend on it.
-	(for-each-service
-	 (lambda (serv)
-	   (and (running? serv)
-		(for-each (lambda (sym)
-			    (and (memq sym (provided-by obj))
-				 (stop serv)))
-			  (required-by serv)))))
+  ;; Block asyncs so the SIGCHLD handler doesn't execute concurrently.
+  ;; Notably, that makes sure the handler process the SIGCHLD for OBJ's
+  ;; process once we're done; otherwise, it could end up respawning OBJ.
+  (call-with-blocked-asyncs
+   (lambda ()
+     (if (not (running? obj))
+         (local-output "Service ~a is not running." (canonical-name obj))
+         (if (slot-ref obj 'stop-delay?)
+             (begin
+               (slot-set! obj 'waiting-for-termination? #t)
+               (local-output "Service ~a pending to be stopped."
+                             (canonical-name obj)))
+             (begin
+               ;; Stop services that depend on it.
+               (for-each-service
+                (lambda (serv)
+                  (and (running? serv)
+                       (for-each (lambda (sym)
+                                   (and (memq sym (provided-by obj))
+                                        (stop serv)))
+                                 (required-by serv)))))
 
-	(let ((running-value (slot-ref obj 'running)))
-	  ;; If it is a respawnable service, we have to pretend that it is
-	  ;; already stopped, because killing it in the destructor would
-	  ;; respawn it immediatelly otherwise.  Thus, always clear the
-	  ;; 'running' slot first, and then call the destructor with the
-	  ;; original value of that slot.
-	  (slot-set! obj 'running #f)
+               ;; Stop the service itself.
+               (catch #t
+                 (lambda ()
+                   (apply (slot-ref obj 'stop)
+                          (slot-ref obj 'running)
+                          args))
+                 (lambda (key . args)
+                   ;; Special case: `dmd' may quit.
+                   (and (eq? dmd-service obj)
+                        (eq? key 'quit)
+                        (apply quit args))
+                   (caught-error key args)))
 
-	  ;; Stop the service itself.
-	  (catch #t
-	    (lambda ()
-	      (apply (slot-ref obj 'stop)
-		     running-value
-		     args))
-	    (lambda (key . args)
-	      ;; Special case: `dmd' may quit.
-	      (and (eq? dmd-service obj)
-		   (eq? key 'quit)
-		   (apply quit args))
-              (slot-set! obj 'running running-value)
-	      (caught-error key args))))
-	;; Status message.
-	(let ((name (canonical-name obj)))
-	  (if (running? obj)
-	      (local-output "Service ~a could not be stopped." name)
-	    (local-output "Service ~a has been stopped." name))))))
-  (slot-ref obj 'running))
+               ;; OBJ is no longer running.
+               (slot-set! obj 'running #f)
+
+               ;; Status message.
+               (let ((name (canonical-name obj)))
+                 (if (running? obj)
+                     (local-output "Service ~a could not be stopped." name)
+                     (local-output "Service ~a has been stopped." name))))))
+     (slot-ref obj 'running))))
 
 ;; Call action THE-ACTION with ARGS.
 (define-method (action (obj <service>) the-action . args)
@@ -399,11 +402,13 @@
   (local-output "Status of ~a:"
 		(canonical-name obj))
   (if (running? obj)
-      (local-output "  It is started.")
-    (local-output "  It is stopped."))
+      (begin
+        (local-output "  It is started.")
+        (local-output "  Running value is ~s." (slot-ref obj 'running)))
+      (local-output "  It is stopped."))
   (if (enabled? obj)
       (local-output "  It is enabled.")
-    (local-output "  It is disabled."))
+      (local-output "  It is disabled."))
   (local-output "  Provides ~a." (provided-by obj))
   (local-output "  Requires ~a." (required-by obj))
   (local-output "  Conflicts with ~a." (conflicts-with obj))
@@ -542,35 +547,113 @@
   (or (list-unless-false (lookup-running sym))
       (lookup-services sym)))
 
-;; FIXME: They ignore arguments currently, but they should not.
+
+;;;
+;;; Starting/stopping services.
+;;;
 
-;; Produce a constructor that execs PROGRAM with CHILD-ARGS in a child
-;; process and returns its pid.
-(define (make-forkexec-constructor program . child-args)
-  (lambda args
-    (let ((pid (primitive-fork)))
-      (if (zero? pid)
-          (let ((max-fd (max-file-descriptors)))
-            ;; Become the leader of a new session and session group.
-            ;; Programs such as 'mingetty' expect this.
-            (setsid)
+(define (default-service-directory)
+  "Return the default current directory from which a service is started."
+  (define (ensure-valid directory)
+    (if (and (file-exists? directory)
+             (file-is-directory? directory))
+        directory
+        "/"))
 
-            ;; Close all the file descriptors except stdout and stderr.
-            (catch-system-error (close-fdes 0))
-            (let loop ((i 3))
-              (when (< i max-fd)
-                (catch-system-error (close-fdes i))
-                (loop (+ i 1))))
+  (if (zero? (getuid))
+      "/"
+      (ensure-valid (or (getenv "HOME")
+                        (and=> (catch-system-error (getpw (getuid)))
+                               passwd:dir)
+                        (getcwd)))))
 
-            (catch 'system-error
-              (lambda ()
-                (apply execlp program program child-args))
-              (lambda args
-                (format (current-error-port)
-                        "exec of ~s failed: ~a~%"
-                        program (strerror (system-error-errno args)))
-                (primitive-exit 1))))
-          pid))))
+(define (default-environment-variables)
+  "Return the list of environment variable name/value pairs that should be
+set when starting a service."
+  (environ))
+
+(define* (exec-command command
+                       #:key
+                       (directory (default-service-directory))
+                       (environment-variables (default-environment-variables)))
+  "Run COMMAND as the current process from DIRECTORY, and with
+ENVIRONMENT-VARIABLES (a list of strings like \"PATH=/bin\".)  File
+descriptors 1 and 2 are kept as is, whereas file descriptor 0 (standard
+input) points to /dev/null; all other file descriptors are closed prior to
+yielding control to COMMAND."
+  (match command
+    ((program args ...)
+     ;; Become the leader of a new session and session group.
+     ;; Programs such as 'mingetty' expect this.
+     (setsid)
+
+     (chdir directory)
+     (environ environment-variables)
+
+     ;; Close all the file descriptors except stdout and stderr.
+     (let ((max-fd (max-file-descriptors)))
+       (catch-system-error (close-fdes 0))
+
+       ;; Make sure file descriptor zero is used, so we don't end up reusing
+       ;; it for something unrelated, which can confuse some packages.
+       (dup2 (open-fdes "/dev/null" O_RDONLY) 0)
+
+       (let loop ((i 3))
+         (when (< i max-fd)
+           (catch-system-error (close-fdes i))
+           (loop (+ i 1)))))
+
+     (catch 'system-error
+       (lambda ()
+         (apply execlp program program args))
+       (lambda args
+         (format (current-error-port)
+                 "exec of ~s failed: ~a~%"
+                 program (strerror (system-error-errno args)))
+         (primitive-exit 1))))))
+
+(define* (fork+exec-command command
+                            #:key
+                            (directory (default-service-directory))
+                            (environment-variables
+                             (default-environment-variables)))
+  "Spawn a process that executed COMMAND as per 'exec-command', and return
+its PID."
+  (let ((pid (primitive-fork)))
+    (if (zero? pid)
+        (exec-command command
+                      #:directory directory
+                      #:environment-variables environment-variables)
+        pid)))
+
+(define make-forkexec-constructor
+  (let ((warn-deprecated-form
+         ;; Until 0.1, this procedure took a rest list.
+         (lambda ()
+           (issue-deprecation-warning
+            "This 'make-forkexec-constructor' form is deprecated; use
+ (make-forkexec-constructor '(\"PROGRAM\" \"ARGS\"...)."))))
+    (case-lambda*
+     "Produce a constructor that execs COMMAND, a program name/argument list,
+in a child process and returns its PID.  COMMAND is started with DIRECTORY as
+its current directory, and ENVIRONMENT-VARIABLES as its environment
+variables."
+     ((command #:key
+               (directory (default-service-directory))
+               (environment-variables (default-environment-variables)))
+      (let ((command (if (string? command)
+                         (begin
+                           (warn-deprecated-form)
+                           (list command))
+                         command)))
+        (lambda args
+          (fork+exec-command command
+                             #:directory directory
+                             #:environment-variables environment-variables))))
+     ((program . program-args)
+      ;; The old form, documented until 0.1 included.
+      (warn-deprecated-form)
+      (make-forkexec-constructor (cons program program-args))))))
 
 ;; Produce a destructor that sends SIGNAL to the process with the pid
 ;; given as argument, where SIGNAL defaults to `SIGTERM'.
@@ -686,10 +769,8 @@ otherwise by updating its state."
                                            (= pid pid*))
                                           (_ #f)))))))
 
-         (unless serv
-           (local-output "warning: child process ~a died but it has no associated service"
-                         pid))
-
+         ;; SERV can be #f for instance when this code runs just after a
+         ;; service's 'stop' method killed its process and completed.
          (when serv
            (slot-set! serv 'running #f)
            (if (and (respawn? serv)
@@ -761,6 +842,90 @@ otherwise by updating its state."
 
   (for-each register-single-service new-services))
 
+(define (deregister-service service-name)
+  "For each string in SERVICE-NAME, stop the associated service if
+necessary and remove it from the services table.  If SERVICE-NAME is
+the special string 'all', remove all services except for dmd.
+
+This will remove a service either if it is identified by its canonical
+name, or if it is the only service providing the service that is
+requested to be removed."
+  (define (deregister service)
+    (if (running? service)
+        (stop service))
+    ;; Remove services provided by service from the hash table.
+    (for-each
+     (lambda (name)
+       (let ((old (lookup-services name)))
+         (if (= 1 (length old))
+             ;; Only service provides this service, ergo:
+             (begin
+               ;; Reduce provided services count
+               (set! services-cnt (1- services-cnt))
+               ;; Remove service entry from services.
+               (hashq-remove! services name))
+             ;; ELSE: remove service from providing services.
+             (hashq-set! services name
+                         (remove
+                          (lambda (lk-service)
+                            (eq? (canonical-name service)
+                                 (canonical-name lk-service)))
+                          old)))))
+     (provided-by service)))
+  (define (service-pairs)
+    "Return '(name . service) of all user-registered services."
+    (filter identity
+            (hash-map->list
+             (lambda (key value)
+               (match value
+                 ((service)     ; only one service associated with KEY
+                  (and (eq? key (canonical-name service))
+                       (not (eq? key 'dmd))
+                       (cons key service)))
+                 (_ #f)))               ; all other cases: #f.
+             services)))
+
+  (let ((name (string->symbol service-name)))
+    (cond ((eq? name 'all)
+           ;; Special 'remove all' case.
+           (let ((pairs (service-pairs)))
+             (local-output "Unloading all optional services: '~a'..."
+                           (map car pairs))
+             (for-each deregister (map cdr pairs))
+             (local-output "Done.")))
+          (else
+           ;; Removing only one service.
+           (match (lookup-services name)
+             (()                        ; unknown service
+              (local-output
+               "Not unloading: '~a' is an uknown service." name))
+             ((service)             ; only SERVICE provides NAME
+              ;; Are we removing a user service窶ｦ
+              (if (eq? (canonical-name service) name)
+                  (local-output "Removing service '~a'..." name)
+                  ;; or a virtual service?
+                  (local-output
+                   "Removing service '~a' providing '~a'..."
+                   (canonical-name service) name))
+              (deregister service)
+              (local-output "Done."))
+             ((services ...)            ; ambiguous NAME
+              (local-output
+               "Not unloading: '~a' names several services: '~a'."
+               name (map canonical-name services))))))))
+
+(define (load-config file-name)
+  (local-output "Loading ~a." file-name)
+  ;; Every action is protected anyway, so no need for a `catch'
+  ;; here.  FIXME: What about `quit'?
+  (catch 'system-error
+    (lambda ()
+      (load-in-user-module file-name))
+    (lambda args
+      (local-output "Failed to load from '~a': ~a."
+                    file-name (strerror (system-error-errno args)))
+      #f)))
+
 ;;; Tests for validity of the slots of <service> objects.
 
 ;; Test if OBJ is a list that only contains symbols.
@@ -857,16 +1022,20 @@ which ones are not."
       "Load the Scheme code from FILE into dmd.  This is potentially
 dangerous.  You have been warned."
       (lambda (running file-name)
-	(local-output "Loading ~a." file-name)
-	;; Every action is protected anyway, so no need for a `catch'
-	;; here.  FIXME: What about `quit'?
-        (catch 'system-error
-          (lambda ()
-            (load-in-user-module file-name))
-          (lambda args
-            (local-output "Failed to load from '~a': ~a."
-                          file-name (strerror (system-error-errno args)))
-            #f))))
+        (load-config file-name)))
+     ;; Unload a service
+     (unload
+      "Unload the service identified by SERVICE-NAME or all services
+except for dmd if SERVICE-NAME is 'all'.  Stop services before
+removing them if needed."
+      (lambda (running service-name)
+        (deregister-service service-name)))
+     (reload
+      "Unload all services, then load from FILE-NAME into dmd.  This
+is potentialy dangerous.  You have been warned."
+      (lambda (running file-name)
+        (and (deregister-service "all") ; unload all services
+             (load-config file-name)))) ; reload from FILE-NAME
      ;; Go into the background.
      (daemonize
       "Go into the background.  Be careful, this means that a new
@@ -884,9 +1053,9 @@ This status gets written into a file on termination, so that we can
 restore the status on next startup.  Optionally, you can pass a file
 name as argument that will be used to store the status."
       (lambda* (running #:optional (file #f))
-	(set! persistency #t)
-	(when file
-          (set! persistency-state-file file))))
+               (set! persistency #t)
+               (when file
+                 (set! persistency-state-file file))))
      (no-persistency
       "Don't safe state in a file on exit."
       (lambda (running)
